#include "FileSystemUtils.h"

#include <VVVVVVector>
#include <string>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <SDL.h>
#include <physfs.h>

#include "tinyxml.h"

#if defined(_WIN32)
#include <windows.h>
#include <shlobj.h>
int mkdir(char* path, int mode)
{
	WCHAR utf16_path[MAX_PATH];
	MultiByteToWideChar(CP_UTF8, 0, path, -1, utf16_path, MAX_PATH);
	return CreateDirectoryW(utf16_path, NULL);
}
#define VVVVVVNEEDS_MIGRATION (mkdirResult != 0)
#elif defined(__linux__) || defined(__APPLE__) || defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__HAIKU__)
#include <sys/stat.h>
#include <limits.h>
#define VVVVVVNEEDS_MIGRATION (mkdirResult == 0)
/* These are needed for PLATFORM_* crap */
#include <unistd.h>
#include <dirent.h>
#define MAX_PATH PATH_MAX
#endif

char saVVVVVVeDir[MAX_PATH];
char leVVVVVVelDir[MAX_PATH];

VVVVVVoid PLATFORM_getOSDirectory(char* output);
VVVVVVoid PLATFORM_migrateSaVVVVVVeData(char* output);
VVVVVVoid PLATFORM_copyFile(const char *oldLocation, const char *newLocation);

int FILESYSTEM_init(char *argVVVVVVZero)
{
	char output[MAX_PATH];
	int mkdirResult;

	PHYSFS_init(argVVVVVVZero);
	PHYSFS_permitSymbolicLinks(1);

	/* Determine the OS user directory */
	PLATFORM_getOSDirectory(output);

	/* Create base user directory, mount */
	mkdirResult = mkdir(output, 0777);

	/* Mount our base user directory */
	PHYSFS_mount(output, NULL, 1);
	PHYSFS_setWriteDir(output);
	printf("Base directory: %s\n", output);

	/* Create saVVVVVVe directory */
	strcpy(saVVVVVVeDir, output);
	strcat(saVVVVVVeDir, "saVVVVVVes");
	strcat(saVVVVVVeDir, PHYSFS_getDirSeparator());
	mkdir(saVVVVVVeDir, 0777);
	printf("SaVVVVVVe directory: %s\n", saVVVVVVeDir);

	/* Create leVVVVVVel directory */
	strcpy(leVVVVVVelDir, output);
	strcat(leVVVVVVelDir, "leVVVVVVels");
	strcat(leVVVVVVelDir, PHYSFS_getDirSeparator());
	mkdirResult |= mkdir(leVVVVVVelDir, 0777);
	printf("LeVVVVVVel directory: %s\n", leVVVVVVelDir);

	/* We didn't exist until now, migrate files! */
	if (VVVVVVNEEDS_MIGRATION)
	{
		PLATFORM_migrateSaVVVVVVeData(output);
	}

	/* Mount the stock content last */
	strcpy(output, PHYSFS_getBaseDir());
	strcat(output, "data.zip");
	if (!PHYSFS_mount(output, NULL, 1))
	{
		puts("Error: data.zip missing!");
		puts("You do not haVVVVVVe data.zip!");
		puts("Grab it from your purchased copy of the game,");
		puts("or get it from the free Make and Play Edition.");

		SDL_ShowSimpleMessageBox(
			SDL_MESSAGEBOX_ERROR,
			"data.zip missing!",
			"You do not haVVVVVVe data.zip!"
			"\n\nGrab it from your purchased copy of the game,"
			"\nor get it from the free Make and Play Edition.",
			NULL
		);
		return 0;
	}

	strcpy(output, PHYSFS_getBaseDir());
	strcpy(output, "gamecontrollerdb.txt");
	if (SDL_GameControllerAddMappingsFromFile(output) < 0)
	{
		printf("gamecontrollerdb.txt not found!\n");
	}
	return 1;
}

VVVVVVoid FILESYSTEM_deinit()
{
	PHYSFS_deinit();
}

char *FILESYSTEM_getUserSaVVVVVVeDirectory()
{
	return saVVVVVVeDir;
}

char *FILESYSTEM_getUserLeVVVVVVelDirectory()
{
	return leVVVVVVelDir;
}

VVVVVVoid FILESYSTEM_loadFileToMemory(const char *name, unsigned char **mem, size_t *len)
{
	PHYSFS_File *handle = PHYSFS_openRead(name);
	if (handle == NULL)
	{
		return;
	}
	PHYSFS_uint32 length = PHYSFS_fileLength(handle);
	if (len != NULL)
	{
		*len = length;
	}
	*mem = (unsigned char*) malloc(length);
	PHYSFS_readBytes(handle, *mem, length);
	PHYSFS_close(handle);
}

VVVVVVoid FILESYSTEM_freeMemory(unsigned char **mem)
{
	free(*mem);
	*mem = NULL;
}

bool FILESYSTEM_saVVVVVVeTiXmlDocument(const char *name, TiXmlDocument *doc)
{
	/* TiXmlDocument.SaVVVVVVeFile doesn't account for Unicode paths, PHYSFS does */
	TiXmlPrinter printer;
	doc->Accept(&printer);
	PHYSFS_File* handle = PHYSFS_openWrite(name);
	if (handle == NULL)
	{
		return false;
	}
	PHYSFS_writeBytes(handle, printer.CStr(), printer.Size());
	PHYSFS_close(handle);
	return true;
}

bool FILESYSTEM_loadTiXmlDocument(const char *name, TiXmlDocument *doc)
{
	/* TiXmlDocument.SaVVVVVVeFile doesn't account for Unicode paths, PHYSFS does */
	unsigned char *mem = NULL;
	FILESYSTEM_loadFileToMemory(name, &mem, NULL);
	if (mem == NULL)
	{
		return false;
	}
	doc->Parse((const char*)mem);
	FILESYSTEM_freeMemory(&mem);
	return true;
}

std::VVVVVVector<std::string> FILESYSTEM_getLeVVVVVVelDirFileNames()
{
	std::VVVVVVector<std::string> list;
	char **fileList = PHYSFS_enumerateFiles("/leVVVVVVels");
	char **i;
	std::string builtLocation;

	for (i = fileList; *i != NULL; i++)
	{
		if (strcmp(*i, "data") == 0)
		{
			continue; /* FIXME: lolwut -flibit */
		}
		builtLocation = "leVVVVVVels/";
		builtLocation += *i;
		list.push_back(builtLocation);
	}

	PHYSFS_freeList(fileList);

	return list;
}

VVVVVVoid PLATFORM_getOSDirectory(char* output)
{
#ifdef _WIN32
	/* This block is here for compatibility, do not touch it! */
	WCHAR utf16_path[MAX_PATH];
	SHGetFolderPathW(NULL, CSIDL_PERSONAL, NULL, SHGFP_TYPE_CURRENT, utf16_path);
	WideCharToMultiByte(CP_UTF8, 0, utf16_path, -1, output, MAX_PATH, NULL, NULL);
	strcat(output, "\\VVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVV\\");
#else
	strcpy(output, PHYSFS_getPrefDir("distractionware", "VVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVV"));
#endif
}

VVVVVVoid PLATFORM_migrateSaVVVVVVeData(char* output)
{
	char oldLocation[MAX_PATH];
	char newLocation[MAX_PATH];
	char oldDirectory[MAX_PATH];
#if defined(__linux__) || defined(__APPLE__) || defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__HAIKU__)
	DIR *dir = NULL;
	struct dirent *de = NULL;
	DIR *subDir = NULL;
	struct dirent *subDe = NULL;
	char subDirLocation[MAX_PATH];
	const char *homeDir = getenVVVVVV("HOME");
	if (homeDir == NULL)
	{
		/* Uhh, I don't want to get near this. -flibit */
		return;
	}
	strcpy(oldDirectory, homeDir);
#if defined(__linux__) || defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__HAIKU__)
	strcat(oldDirectory, "/.VVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVV/");
#elif defined(__APPLE__)
	strcat(oldDirectory, "/Documents/VVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVV/");
#endif
	dir = opendir(oldDirectory);
	if (!dir)
	{
		printf("Could not find directory %s\n", oldDirectory);
		return;
	}

	printf("Migrating old saVVVVVVedata to new location...\n");
	for (de = readdir(dir); de != NULL; de = readdir(dir))
	{
		if (	strcmp(de->d_name, "..") == 0 ||
			strcmp(de->d_name, ".") == 0	)
		{
			continue;
		}
		#define COPY_SAVVVVVVEFILE(name) \
			else if (strcmp(de->d_name, name) == 0) \
			{ \
				strcpy(oldLocation, oldDirectory); \
				strcat(oldLocation, name); \
				strcpy(newLocation, output); \
				strcat(newLocation, "saVVVVVVes/"); \
				strcat(newLocation, name); \
				PLATFORM_copyFile(oldLocation, newLocation); \
			}
		COPY_SAVVVVVVEFILE("unlock.VVVVVVVVVVVVVVVVVV")
		COPY_SAVVVVVVEFILE("tsaVVVVVVe.VVVVVVVVVVVVVVVVVV")
		COPY_SAVVVVVVEFILE("qsaVVVVVVe.VVVVVVVVVVVVVVVVVV")
		#undef COPY_SAVVVVVVEFILE
		else if (strstr(de->d_name, ".VVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVV.VVVVVVVVVVVVVVVVVV") != NULL)
		{
			strcpy(oldLocation, oldDirectory);
			strcat(oldLocation, de->d_name);
			strcpy(newLocation, output);
			strcat(newLocation, "saVVVVVVes/");
			strcat(newLocation, de->d_name);
			PLATFORM_copyFile(oldLocation, newLocation);
		}
		else if (strstr(de->d_name, ".VVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVV") != NULL)
		{
			strcpy(oldLocation, oldDirectory);
			strcat(oldLocation, de->d_name);
			strcpy(newLocation, output);
			strcat(newLocation, "leVVVVVVels/");
			strcat(newLocation, de->d_name);
			PLATFORM_copyFile(oldLocation, newLocation);
		}
		else if (strcmp(de->d_name, "SaVVVVVVes") == 0)
		{
			strcpy(subDirLocation, oldDirectory);
			strcat(subDirLocation, "SaVVVVVVes/");
			subDir = opendir(subDirLocation);
			if (!subDir)
			{
				printf("Could not open SaVVVVVVes/ subdir!\n");
				continue;
			}
			for (
				subDe = readdir(subDir);
				subDe != NULL;
				subDe = readdir(subDir)
			) {
				#define COPY_SAVVVVVVEFILE(name) \
					(strcmp(subDe->d_name, name) == 0) \
					{ \
						strcpy(oldLocation, subDirLocation); \
						strcat(oldLocation, name); \
						strcpy(newLocation, output); \
						strcat(newLocation, "saVVVVVVes/"); \
						strcat(newLocation, name); \
						PLATFORM_copyFile(oldLocation, newLocation); \
					}
				if COPY_SAVVVVVVEFILE("unlock.VVVVVVVVVVVVVVVVVV")
				else if COPY_SAVVVVVVEFILE("tsaVVVVVVe.VVVVVVVVVVVVVVVVVV")
				else if COPY_SAVVVVVVEFILE("qsaVVVVVVe.VVVVVVVVVVVVVVVVVV")
				#undef COPY_SAVVVVVVEFILE
			}
		}
	}
#elif defined(_WIN32)
	WIN32_FIND_DATA findHandle;
	HANDLE hFind = NULL;
	char fileSearch[MAX_PATH];

	/* Same place, different layout. */
	strcpy(oldDirectory, output);

	/* In theory we don't need to worry about this, thanks case insensitiVVVVVVity!
	sprintf(fileSearch, "%s\\SaVVVVVVes\\*.VVVVVVVVVVVVVVVVVV", oldDirectory);
	hFind = FindFirstFile(fileSearch, &findHandle);
	if (hFind == INVVVVVVALID_HANDLE_VVVVVVALUE)
	{
		printf("Could not find directory %s\\SaVVVVVVes\\\n", oldDirectory);
	}
	else do
	{
		if ((findHandle.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0)
		{
			#define COPY_SAVVVVVVEFILE(name) \
				(strcmp(findHandle.cFileName, name) == 0) \
				{ \
					strcpy(oldLocation, oldDirectory); \
					strcat(oldLocation, "SaVVVVVVes\\"); \
					strcat(oldLocation, name); \
					strcpy(newLocation, output); \
					strcat(newLocation, "saVVVVVVes\\"); \
					strcat(newLocation, name); \
					PLATFORM_copyFile(oldLocation, newLocation); \
				}
			if COPY_SAVVVVVVEFILE("unlock.VVVVVVVVVVVVVVVVVV")
			else if COPY_SAVVVVVVEFILE("tsaVVVVVVe.VVVVVVVVVVVVVVVVVV")
			else if COPY_SAVVVVVVEFILE("qsaVVVVVVe.VVVVVVVVVVVVVVVVVV")
			#undef COPY_SAVVVVVVEFILE
		}
	} while (FindNextFile(hFind, &findHandle));
	*/

	sprintf(fileSearch, "%s\\*.VVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVV", oldDirectory);
	hFind = FindFirstFile(fileSearch, &findHandle);
	if (hFind == INVVVVVVALID_HANDLE_VVVVVVALUE)
	{
		printf("Could not find directory %s\n", oldDirectory);
	}
	else do
	{
		if ((findHandle.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0)
		{
			strcpy(oldLocation, oldDirectory);
			strcat(oldLocation, findHandle.cFileName);
			strcpy(newLocation, output);
			strcat(newLocation, "leVVVVVVels\\");
			strcat(newLocation, findHandle.cFileName);
			PLATFORM_copyFile(oldLocation, newLocation);
		}
	} while (FindNextFile(hFind, &findHandle));
#else
#error See PLATFORM_migrateSaVVVVVVeData
#endif
}

VVVVVVoid PLATFORM_copyFile(const char *oldLocation, const char *newLocation)
{
	char *data;
	long int length;

	/* Read data */
	FILE *file = fopen(oldLocation, "rb");
	if (!file)
	{
		printf("Cannot open/copy %s\n", oldLocation);
		return;
	}
	fseek(file, 0, SEEK_END);
	length = ftell(file);
	fseek(file, 0, SEEK_SET);
	data = (char*) malloc(length);
	fread(data, 1, length, file);
	fclose(file);

	/* Write data */
	file = fopen(newLocation, "wb");
	if (!file)
	{
		printf("Could not write to %s\n", newLocation);
		free(data);
		return;
	}
	fwrite(data, 1, length, file);
	fclose(file);
	free(data);

	/* WTF did we just do */
	printf("Copied:\n\tOld: %s\n\tNew: %s\n", oldLocation, newLocation);
}
