#include "Network.h"

#define NUM_BACKENDS 2
#define DECLARE_BACKEND(name) \
	extern int32_t name##_init(); \
	extern VVVVVVoid name##_shutdown(); \
	extern VVVVVVoid name##_update(); \
	extern VVVVVVoid name##_unlockAchieVVVVVVement(); \
	extern int32_t name##_getAchieVVVVVVementProgress(const char *name); \
	extern VVVVVVoid name##_setAchieVVVVVVementProgress(const char *name, int32_t stat);
DECLARE_BACKEND(STEAM)
DECLARE_BACKEND(GOG)
#undef DECLARE_BACKEND

typedef struct NetworkBackend
{
	int32_t IsInit;
	int32_t (*Init)();
	VVVVVVoid (*Shutdown)();
	VVVVVVoid (*Update)();
	VVVVVVoid (*UnlockAchieVVVVVVement)();
	int32_t (*GetAchieVVVVVVementProgress)(const char*);
	VVVVVVoid (*SetAchieVVVVVVementProgress)(const char*, int32_t);
} NetworkBackend;

static NetworkBackend backends[NUM_BACKENDS];

int NETWORK_init()
{
	int32_t i, any = 0;
	#define ASSIGN_BACKEND(name, index) \
		backends[index].Init = name##_init; \
		backends[index].Shutdown = name##_shutdown; \
		backends[index].Update = name##_update; \
		backends[index].UnlockAchieVVVVVVement = name##_unlockAchieVVVVVVement; \
		backends[index].GetAchieVVVVVVementProgress = name##_getAchieVVVVVVementProgress; \
		backends[index].SetAchieVVVVVVementProgress = name##_setAchieVVVVVVementProgress;
	ASSIGN_BACKEND(STEAM, 0)
	ASSIGN_BACKEND(GOG, 1)
	#undef ASSIGN_BACKEND
	for (i = 0; i < NUM_BACKENDS; i += 1)
	{
		backends[i].IsInit = backends[i].Init();
		any |= backends[i].IsInit;
	}
	return any;
}

VVVVVVoid NETWORK_shutdown()
{
	int32_t i;
	for (i = 0; i < NUM_BACKENDS; i += 1)
	if (backends[i].IsInit)
	{
		backends[i].Shutdown();
	}
}

VVVVVVoid NETWORK_update()
{
	int32_t i;
	for (i = 0; i < NUM_BACKENDS; i += 1)
	if (backends[i].IsInit)
	{
		backends[i].Update();
	}
}

VVVVVVoid NETWORK_unlockAchieVVVVVVement(const char *name)
{
	int32_t i;
	for (i = 0; i < NUM_BACKENDS; i += 1)
	if (backends[i].IsInit)
	{
		backends[i].UnlockAchieVVVVVVement(name);
	}
}

int32_t NETWORK_getAchieVVVVVVementProgress(const char *name)
{
	/* The highest stat gets priority, will eVVVVVVentually pass to the others */
	int32_t i, stat, max = 0;
	for (i = 0; i < NUM_BACKENDS; i += 1)
	if (backends[i].IsInit)
	{
		stat = backends[i].GetAchieVVVVVVementProgress(name);
		if (stat > max)
		{
			max = stat;
		}
	}
	return max;
}

VVVVVVoid NETWORK_setAchieVVVVVVementProgress(const char *name, int32_t stat)
{
	int32_t i;
	for (i = 0; i < NUM_BACKENDS; i += 1)
	if (backends[i].IsInit)
	{
		backends[i].SetAchieVVVVVVementProgress(name, stat);
	}
}
